"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.hasFacetCodes = exports.throwOdspNetworkError = exports.enrichOdspError = exports.createOdspNetworkError = exports.parseFacetCodes = exports.tryParseErrorResponse = exports.OdspRedirectError = exports.getSPOAndGraphRequestIdsFromResponse = exports.OdspServiceReadOnlyErrorCode = exports.fetchIncorrectResponse = void 0;
const internal_1 = require("@fluidframework/driver-definitions/internal");
const internal_2 = require("@fluidframework/driver-utils/internal");
const internal_3 = require("@fluidframework/odsp-driver-definitions/internal");
const internal_4 = require("@fluidframework/telemetry-utils/internal");
// odsp-doclib-utils and odsp-driver will always release together and share the same pkgVersion
const packageVersion_js_1 = require("./packageVersion.js");
const parseAuthErrorClaims_js_1 = require("./parseAuthErrorClaims.js");
const parseAuthErrorTenant_js_1 = require("./parseAuthErrorTenant.js");
// no response, or can't parse response
/**
 * @internal
 */
exports.fetchIncorrectResponse = 712;
// Error code for when the server state is read only and client tries to write. This code is set by the server
// and is not likely to change.
/**
 * @internal
 */
exports.OdspServiceReadOnlyErrorCode = "serviceReadOnly";
/**
 * @internal
 */
function getSPOAndGraphRequestIdsFromResponse(headers) {
    // We rename headers so that otel doesn't scrub them away. Otel doesn't allow
    // certain characters in headers including '-'
    const headersToLog = [
        { headerName: "sprequestguid", logName: "sprequestguid" },
        { headerName: "request-id", logName: "requestId" },
        { headerName: "client-request-id", logName: "clientRequestId" },
        { headerName: "x-msedge-ref", logName: "xMsedgeRef" },
        { headerName: "X-Fluid-Retries", logName: "serverRetries" },
        { headerName: "content-encoding", logName: "contentEncoding" },
        { headerName: "content-type", logName: "contentType" },
    ];
    const additionalProps = {
        sprequestduration: (0, internal_4.numberFromString)(headers.get("sprequestduration")),
        contentsize: (0, internal_4.numberFromString)(headers.get("content-length")),
    };
    headersToLog.forEach((header) => {
        const headerValue = headers.get(header.headerName);
        if (headerValue !== undefined && headerValue !== null) {
            additionalProps[header.logName] = headerValue;
        }
    });
    // x-fluid-telemetry contains a key value pair in the following format:
    // x-fluid-telemetry:key1=value1,key2,key3=value3,
    // Ex. x-fluid-telemetry:Origin=c
    const fluidTelemetry = headers.get("x-fluid-telemetry");
    if (fluidTelemetry !== undefined && fluidTelemetry !== null) {
        additionalProps.xFluidTelemetry = fluidTelemetry;
        const keyValueMap = fluidTelemetry
            .split(",")
            .map((keyValuePair) => keyValuePair.split("="));
        for (const [key, value] of keyValueMap) {
            if ("Origin" === key.trim()) {
                let fieldValue;
                switch (value?.trim()) {
                    case "c":
                        fieldValue = "cache";
                        break;
                    case "g":
                        fieldValue = "graph";
                        break;
                    default:
                        fieldValue = value?.trim();
                }
                const logName = "responseOrigin";
                additionalProps[logName] = fieldValue;
                break;
            }
        }
    }
    return additionalProps;
}
exports.getSPOAndGraphRequestIdsFromResponse = getSPOAndGraphRequestIdsFromResponse;
/**
 * Error encapsulating the error response from ODSP containing the redirect location when a resource has moved
 *
 * @internal
 */
class OdspRedirectError extends internal_4.LoggingError {
    constructor(message, redirectLocation, props) {
        // do not log redirectLocation (URL can contain sensitive info)
        super(message, props, new Set(["redirectLocation"]));
        this.redirectLocation = redirectLocation;
        this.errorType = internal_1.DriverErrorTypes.fileNotFoundOrAccessDeniedError;
        this.canRetry = false;
    }
}
exports.OdspRedirectError = OdspRedirectError;
/** Empirically-based type guard for error responses from ODSP */
function isOdspErrorResponse(x) {
    const error = x?.error;
    return (typeof error?.message === "string" &&
        (error?.code === undefined || typeof error?.code === "string"));
}
/**
 * @internal
 */
function tryParseErrorResponse(response) {
    try {
        if (response !== undefined) {
            const parsed = JSON.parse(response);
            if (isOdspErrorResponse(parsed)) {
                return { success: true, errorResponse: parsed };
            }
        }
    }
    catch (e) { }
    return { success: false };
}
exports.tryParseErrorResponse = tryParseErrorResponse;
/**
 * @internal
 */
function parseFacetCodes(errorResponse) {
    const stack = [];
    let error = errorResponse.error;
    while (typeof error === "object" && error !== null) {
        if (error.code !== undefined) {
            stack.unshift(error.code);
        }
        error = error.innerError;
    }
    return stack;
}
exports.parseFacetCodes = parseFacetCodes;
/**
 * @internal
 */
function createOdspNetworkError(errorMessage, statusCode, retryAfterSeconds, response, responseText, props = {}) {
    let error;
    const parseResult = tryParseErrorResponse(responseText);
    let facetCodes;
    let innerMostErrorCode;
    if (parseResult.success) {
        const errorResponse = parseResult.errorResponse;
        // logging the error response message
        props.responseMessage = errorResponse.error.message;
        facetCodes = parseFacetCodes(errorResponse);
        if (facetCodes !== undefined) {
            innerMostErrorCode = facetCodes[0];
            props.innerMostErrorCode = innerMostErrorCode;
        }
    }
    let redirectLocation;
    const driverProps = { driverVersion: packageVersion_js_1.pkgVersion, statusCode, ...props };
    switch (statusCode) {
        case 400:
            if (innerMostErrorCode === "fluidInvalidSchema") {
                error = new internal_2.FluidInvalidSchemaError(errorMessage, driverProps);
                break;
            }
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.genericNetworkError, driverProps);
            break;
        case 401:
        case 403:
            // The server throws 403 status code with innerMostError code as "serviceReadOnly" for cases where the
            // database on server becomes readonly. The driver retries for such cases with exponential backup logic.
            if (innerMostErrorCode === exports.OdspServiceReadOnlyErrorCode) {
                error = new internal_2.RetryableError(errorMessage, internal_3.OdspErrorTypes.serviceReadOnly, driverProps);
            }
            else if (innerMostErrorCode === "blockedIPAddress" ||
                innerMostErrorCode === "conditionalAccessPolicyEnforced") {
                error = new internal_2.NonRetryableError("IP Address is blocked", internal_3.OdspErrorTypes.blockedIPAddress, driverProps);
            }
            else {
                const claims = response?.headers
                    ? (0, parseAuthErrorClaims_js_1.parseAuthErrorClaims)(response.headers)
                    : undefined;
                const tenantId = response?.headers
                    ? (0, parseAuthErrorTenant_js_1.parseAuthErrorTenant)(response.headers)
                    : undefined;
                error = new internal_2.AuthorizationError(errorMessage, claims, tenantId, driverProps);
            }
            break;
        case 404:
            if (parseResult.success) {
                // The location of file can move on Spo. If the manual redirect prefer header is added to network call
                // it returns 404 error instead of 308. Error thrown by server will contain the new redirect location.
                // For reference we can look here: \packages\drivers\odsp-driver\src\fetchSnapshot.ts
                const responseError = parseResult?.errorResponse?.error;
                redirectLocation = responseError?.["@error.redirectLocation"];
                if (redirectLocation !== undefined) {
                    error = new OdspRedirectError(errorMessage, redirectLocation, driverProps);
                    break;
                }
            }
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.fileNotFoundOrAccessDeniedError, driverProps);
            break;
        case 406:
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.unsupportedClientProtocolVersion, driverProps);
            break;
        case 410:
            error = new internal_2.NonRetryableError(errorMessage, internal_3.OdspErrorTypes.cannotCatchUp, driverProps);
            break;
        case 409:
            // This status code is sent by the server when the client and server epoch mismatches.
            // The client sets its epoch version in the calls it makes to the server and if that mismatches
            // with the server epoch version, the server throws this error code.
            // This indicates that the file/container has been modified externally.
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.fileOverwrittenInStorage, driverProps);
            break;
        case 412:
            // "Precondition Failed" error - happens when uploadSummaryWithContext uses wrong parent.
            // Resubmitting same payload is not going to help, so this is non-recoverable failure!
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.genericNetworkError, driverProps);
            break;
        case 413:
            error = new internal_2.NonRetryableError(errorMessage, internal_3.OdspErrorTypes.snapshotTooBig, driverProps);
            break;
        case 414:
            error = new internal_2.NonRetryableError(errorMessage, internal_3.OdspErrorTypes.invalidFileNameError, driverProps);
            break;
        case 423: // File locked
            if (innerMostErrorCode === "resourceLocked" ||
                innerMostErrorCode === "resourceCheckedOut") {
                error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.fileIsLocked, driverProps);
                break;
            }
        case 500:
            error = new internal_2.RetryableError(errorMessage, internal_1.DriverErrorTypes.genericNetworkError, driverProps);
            break;
        case 501:
            error = new internal_2.NonRetryableError(errorMessage, internal_3.OdspErrorTypes.fluidNotEnabled, driverProps);
            break;
        case 507:
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.outOfStorageError, driverProps);
            break;
        case exports.fetchIncorrectResponse:
            // Note that getWithRetryForTokenRefresh will retry it once, then it becomes non-retryable error
            error = new internal_2.NonRetryableError(errorMessage, internal_1.DriverErrorTypes.incorrectServerResponse, driverProps);
            break;
        default:
            const retryAfterMs = retryAfterSeconds !== undefined ? retryAfterSeconds * 1000 : undefined;
            error = (0, internal_2.createGenericNetworkError)(errorMessage, { canRetry: true, retryAfterMs }, driverProps);
            break;
    }
    // Set this to true as createOdspNetworkError is called to handle error response from service.
    error.addTelemetryProperties({ endpointReached: true });
    enrichOdspError(error, response, facetCodes, undefined);
    return error;
}
exports.createOdspNetworkError = createOdspNetworkError;
/**
 * @internal
 */
function enrichOdspError(error, response, facetCodes, props = {}) {
    error.online = internal_2.OnlineStatus[(0, internal_2.isOnline)()];
    if (facetCodes !== undefined) {
        error.facetCodes = facetCodes;
    }
    if (response) {
        props.responseType = response.type;
        if (response.headers) {
            const headers = getSPOAndGraphRequestIdsFromResponse(response.headers);
            for (const key of Object.keys(headers)) {
                props[key] = headers[key];
            }
            error.serverEpoch = response.headers.get("x-fluid-epoch") ?? undefined;
        }
    }
    error.addTelemetryProperties(props);
    return error;
}
exports.enrichOdspError = enrichOdspError;
/**
 * Throws network error - an object with a bunch of network related properties
 * @internal
 */
function throwOdspNetworkError(errorMessage, statusCode, response, responseText, props) {
    const networkError = createOdspNetworkError(errorMessage, statusCode, numberFromHeader(response.headers.get("retry-after")) /* retryAfterSeconds */, response, responseText, props);
    networkError.addTelemetryProperties({ odspError: true, storageServiceError: true });
    // eslint-disable-next-line @typescript-eslint/no-throw-literal
    throw networkError;
}
exports.throwOdspNetworkError = throwOdspNetworkError;
function numberFromHeader(header) {
    if (header === null) {
        return undefined;
    }
    const n = Number(header);
    if (Number.isNaN(n)) {
        return undefined;
    }
    return n;
}
/**
 * @internal
 */
function hasFacetCodes(x) {
    return Array.isArray(x?.facetCodes);
}
exports.hasFacetCodes = hasFacetCodes;
//# sourceMappingURL=odspErrorUtils.js.map